import numpy as np
import pandas as pd
from pandas.core.common import flatten


def empty_df(columns=None):
    columns = [] if columns is None else columns
    empty = {col: [] for col in columns}
    return pd.DataFrame(empty, columns=columns, dtype=np.unicode)


def setwrap(value):
    """
    Returns a flattened and stringified set from the given object or iterable.

    For use in public functions which accept argmuents or kwargs that can be
    one object or a list of objects.
    """
    return set(map(np.unicode, set(flatten([value]))))



try:
    from functools import lru_cache
except ImportError:
    from functools32 import lru_cache

from contextlib import contextmanager
import io
import os
from zipfile import ZipFile

import networkx as nx
import numpy as np
import pandas as pd

from partridge.config import default_config, empty_config
from partridge.utilities import empty_df, setwrap


def read_file(filename):
    return property(lambda feed: feed.get(filename))


class feed(object):
    def __init__(self, path, config=None, view=None):
        self.path = path
        self.is_dir = os.path.isdir(self.path)
        self.config = default_config() if config is None else config
        self.view = {} if view is None else view
        self.zmap = {}

        assert os.path.isfile(self.path) or self.is_dir, \
            'File or path not found: {}'.format(self.path)

        assert nx.is_directed_acyclic_graph(self.config), \
            'Config must be a DAG'

        roots = {n for n, d in self.config.out_degree() if d == 0}
        for filename, param in self.view.items():
            assert filename in roots, \
                'Filter param given for a non-root node ' \
                'of the config graph: {} {}'.format(filename, param)

        if self.is_dir:
            self._verify_folder_contents()
        else:
            self._verify_zip_contents()

    agency = read_file('agency.txt')
    calendar = read_file('calendar.txt')
    calendar_dates = read_file('calendar_dates.txt')
    fare_attributes = read_file('fare_attributes.txt')
    fare_rules = read_file('fare_rules.txt')
    feed_info = read_file('feed_info.txt')
    frequencies = read_file('frequencies.txt')
    routes = read_file('routes.txt')
    shapes = read_file('shapes.txt')
    stops = read_file('stops.txt')
    stop_times = read_file('stop_times.txt')
    transfers = read_file('transfers.txt')
    trips = read_file('trips.txt')

    @lru_cache(maxsize=None)
    def get(self, filename):
        config = self.config

        # Get config for node
        node = config.nodes.get(filename, {})
        columns = node.get('required_columns', [])
        converters = node.get('converters', {})

        # If the file isn't in the zip, return an empty DataFrame.
        if filename not in self.zmap:
            return empty_df(columns)

        # Gather the dependencies between this file and others
        file_dependencies = {
            depfile: data['dependencies'].items()
            for _, depfile, data in config.out_edges(filename, data=True)
            if 'dependencies' in data
        }

        # Gather applicable view filter params
        view_filters = {
            # column name : set of strings
            col: setwrap(values)
            for col, values in self.view.get(filename, {}).items()
        }

        # Read CSV in chunks, prune it according to the dependency graph
        chunks = []
        with self._io_adapter(filename) as iowrapper:
            # Build a chunked DataFrame reader
            reader = pd.read_csv(
                iowrapper, chunksize=10000,
                dtype=np.unicode, index_col=False,
                low_memory=False, skipinitialspace=True)

            # Process the file in chunks
            for i, chunk in enumerate(reader):
                # Cleanup column names just to be safe
                chunk = chunk.rename(columns=lambda x: x.strip())

                if i == 0:
                    # Track the actual columns in the file if present
                    columns = list(chunk.columns)

                # Apply view filters
                for col, values in view_filters.items():
                    # If applicable, filter this chunk by the
                    # given set of values
                    if col in chunk.columns:
                        chunk = chunk[chunk[col].isin(values)]

                # Prune the chunk
                for depfile, dependencies in file_dependencies.items():
                    # Read the filtered, pruned, and cached file dependency
                    depdf = self.get(depfile)

                    for col, depcol in dependencies:
                        # If applicable, prune this chunk by the other
                        if col in chunk.columns and depcol in depdf.columns:
                            chunk = chunk[chunk[col].isin(depdf[depcol])]

                # Discard entirely filtered/pruned chunks
                if not chunk.empty:
                    chunks.append(chunk)

        # If all chunks were completely filtered/pruned away,
        # return an empty DataFrame.
        if len(chunks) == 0:
            return empty_df(columns)

        # Concatenate chunks into one DataFrame
        df = pd.concat(chunks)

        # Apply type conversions, strip leading/trailing whitespace
        for col in df.columns:
            if df[col].any():
                df[col] = df[col].str.strip()
                if col in converters:
                    vfunc = converters[col]
                    df[col] = vfunc(df[col])

        return df

    @contextmanager
    def _io_adapter(self, filename):
        """
        Yield an IO object for the given file
        from a zip file or folder
        """
        if self.is_dir:
            with open(self.zmap[filename], 'rb') as iowrapper:
                yield iowrapper
        else:
            with ZipFile(self.path) as zipreader:
                with zipreader.open(self.zmap[filename], 'r') as zfile:
                    with io.TextIOWrapper(zfile,
                                          encoding='utf-8-sig') as iowrapper:
                        yield iowrapper

    def _verify_zip_contents(self):
        """
        Verify that the folder does not contain multiple files
        of the same name. Load file paths into internal dictionary.
        """
        with ZipFile(self.path) as zipreader:
            for zpath in zipreader.namelist():
                basename = os.path.basename(zpath)
                if zpath.endswith('.txt'):
                    assert basename not in self.zmap, \
                        'More than one {} in zip'.format(basename)
                self.zmap[basename] = zpath

    def _verify_folder_contents(self):
        """
        Verify that the folder does not contain multiple files
        of the same name. Load file paths into internal dictionary.
        """
        files = [
            os.path.join(self.path, f)
            for f in os.listdir(self.path)
            if os.path.isfile(os.path.join(self.path, f))
        ]
        for gtfs_file in files:
            basename = os.path.basename(gtfs_file)
            if gtfs_file.endswith('.txt'):
                assert gtfs_file not in self.zmap, \
                    'More than one {} in zip'.format(basename)
            self.zmap[basename] = gtfs_file


# No pruning or type coercion
class raw_feed(feed):
    def __init__(self, path):
        super(raw_feed, self).__init__(path, config=empty_config())



from partridge.__version__ import __version__
from partridge.gtfs import feed, raw_feed
from partridge.readers import (
    get_representative_feed,
    read_busiest_date,
    read_service_ids_by_date,
    read_dates_by_service_ids,
    read_trip_counts_by_date,
)
from partridge.writers import (
    extract_agencies,
    extract_routes,
)


__all__ = [
    '__version__',
    'feed',
    'raw_feed',
    'get_representative_feed',
    'read_busiest_date',
    'read_service_ids_by_date',
    'read_dates_by_service_ids',
    'read_trip_counts_by_date',
    'extract_agencies',
    'extract_routes',
]



from collections import defaultdict
import datetime

from partridge.gtfs import feed as mkfeed, raw_feed
from partridge.parsers import vparse_date


DAY_NAMES = (
    'monday', 'tuesday', 'wednesday', 'thursday', 'friday',
    'saturday', 'sunday')


'''Public'''


def get_representative_feed(path):
    '''Return a feed filtered to the busiest date'''
    _, service_ids = read_busiest_date(path)
    view = {'trips.txt': {'service_id': service_ids}}
    return mkfeed(path, view=view)


def read_busiest_date(path):
    '''Find the date with the most trips'''
    feed = raw_feed(path)

    service_ids_by_date = _service_ids_by_date(feed)
    trip_counts_by_date = _trip_counts_by_date(feed)

    date, _ = max(trip_counts_by_date.items(), key=lambda p: (p[1], p[0]))
    service_ids = service_ids_by_date[date]

    return date, service_ids


def read_service_ids_by_date(path):
    '''Find all service identifiers by date'''
    feed = raw_feed(path)
    return _service_ids_by_date(feed)


def read_dates_by_service_ids(path):
    '''Find dates with identical service'''
    feed = raw_feed(path)
    return _dates_by_service_ids(feed)


def read_trip_counts_by_date(path):
    '''A useful proxy for busyness'''
    feed = raw_feed(path)
    return _trip_counts_by_date(feed)


'''Private'''


def _service_ids_by_date(feed):
    results = defaultdict(set)
    removals = defaultdict(set)

    service_ids = set(feed.trips.service_id)
    calendar = feed.calendar
    caldates = feed.calendar_dates

    if not calendar.empty:
        # Only consider calendar.txt rows with applicable trips
        calendar = calendar[calendar.service_id.isin(service_ids)].copy()

    if not caldates.empty:
        # Only consider calendar_dates.txt rows with applicable trips
        caldates = caldates[caldates.service_id.isin(service_ids)].copy()

    if not calendar.empty:
        # Parse dates
        calendar.start_date = vparse_date(calendar.start_date)
        calendar.end_date = vparse_date(calendar.end_date)

        # Build up results dict from calendar ranges
        for _, cal in calendar.iterrows():
            start = cal.start_date.toordinal()
            end = cal.end_date.toordinal()

            dow = {i: cal[day] for i, day in enumerate(DAY_NAMES)}
            for ordinal in range(start, end + 1):
                date = datetime.date.fromordinal(ordinal)
                if int(dow[date.weekday()]):
                    results[date].add(cal.service_id)

    if not caldates.empty:
        # Parse dates
        caldates.date = vparse_date(caldates.date)

        # Split out additions and removals
        cdadd = caldates[caldates.exception_type == '1']
        cdrem = caldates[caldates.exception_type == '2']

        # Add to results by date
        for _, cd in cdadd.iterrows():
            results[cd.date].add(cd.service_id)

        # Collect removals
        for _, cd in cdrem.iterrows():
            removals[cd.date].add(cd.service_id)

        # Finally, process removals by date
        for date in removals:
            for service_id in removals[date]:
                if service_id in results[date]:
                    results[date].remove(service_id)

            # Drop the key from results if no service present
            if len(results[date]) == 0:
                del results[date]

    return {k: frozenset(v) for k, v in results.items()}


def _dates_by_service_ids(feed):
    results = defaultdict(set)
    for date, service_ids in _service_ids_by_date(feed).items():
        results[service_ids].add(date)
    return dict(results)


def _trip_counts_by_date(feed):
    results = defaultdict(int)
    trips = feed.trips
    for service_ids, dates in _dates_by_service_ids(feed).items():
        trip_count = trips[trips.service_id.isin(service_ids)].shape[0]
        for date in dates:
            results[date] += trip_count
    return dict(results)



import networkx as nx

from partridge.parsers import \
    vparse_date, \
    vparse_time, \
    vparse_numeric


def empty_config():
    return nx.DiGraph()


'''
Default configs
'''


def default_config():
    G = empty_config()
    add_edge_config(G)
    add_node_config(G)
    return G


def add_edge_config(g):
    g.add_edges_from([
        ('agency.txt', 'routes.txt', {
            'dependencies': {
                'agency_id': 'agency_id',
            },
        }),
        ('calendar.txt', 'trips.txt', {
            'dependencies': {
                'service_id': 'service_id',
            },
        }),
        ('calendar_dates.txt', 'trips.txt', {
            'dependencies': {
                'service_id': 'service_id',
            },
        }),
        ('fare_attributes.txt', 'fare_rules.txt', {
            'dependencies': {
                'fare_id': 'fare_id',
            },
        }),
        ('fare_rules.txt', 'stops.txt', {
            'dependencies': {
                'origin_id': 'zone_id',
                'destination_id': 'zone_id',
                'contains_id': 'zone_id',
            },
        }),
        ('fare_rules.txt', 'routes.txt', {
            'dependencies': {
                'route_id': 'route_id',
            },
        }),
        ('frequencies.txt', 'trips.txt', {
            'dependencies': {
                'trip_id': 'trip_id',
            },
        }),
        ('routes.txt', 'trips.txt', {
            'dependencies': {
                'route_id': 'route_id',
            },
        }),
        ('shapes.txt', 'trips.txt', {
            'dependencies': {
                'shape_id': 'shape_id',
            },
        }),
        ('stops.txt', 'stop_times.txt', {
            'dependencies': {
                'stop_id': 'stop_id',
            },
        }),
        ('stop_times.txt', 'trips.txt', {
            'dependencies': {
                'trip_id': 'trip_id',
            },
        }),
        ('transfers.txt', 'stops.txt', {
            'dependencies': {
                'from_stop_id': 'stop_id',
                'to_stop_id': 'stop_id',
            },
        }),
    ])


def add_node_config(g):
    g.add_nodes_from([
        ('agency.txt', {
            'required_columns': (
                'agency_name',
                'agency_url',
                'agency_timezone',
            ),
        }),
        ('calendar.txt', {
            'converters': {
                'start_date': vparse_date,
                'end_date': vparse_date,
                'monday': vparse_numeric,
                'tuesday': vparse_numeric,
                'wednesday': vparse_numeric,
                'thursday': vparse_numeric,
                'friday': vparse_numeric,
                'saturday': vparse_numeric,
                'sunday': vparse_numeric,
            },
            'required_columns': (
                'service_id',
                'monday',
                'tuesday',
                'wednesday',
                'thursday',
                'friday',
                'saturday',
                'sunday',
                'start_date',
                'end_date',
            ),
        }),
        ('calendar_dates.txt', {
            'converters': {
                'date': vparse_date,
                'exception_type': vparse_numeric,
            },
            'required_columns': (
                'service_id',
                'date',
                'exception_type',
            ),
        }),
        ('fare_attributes.txt', {
            'converters': {
                'price': vparse_numeric,
                'payment_method': vparse_numeric,
                'transfer_duration': vparse_numeric,
            },
            'required_columns': (
                'fare_id',
                'price',
                'currency_type',
                'payment_method',
                'transfers',
            ),
        }),
        ('fare_rules.txt', {
            'required_columns': (
                'fare_id',
            ),
        }),
        ('feed_info.txt', {
            'converters': {
                'feed_start_date': vparse_date,
                'feed_end_date': vparse_date,
            },
            'required_columns': (
                'feed_publisher_name',
                'feed_publisher_url',
                'feed_lang',
            ),
        }),
        ('frequencies.txt', {
            'converters': {
                'headway_secs': vparse_numeric,
                'exact_times': vparse_numeric,
                'start_time': vparse_time,
                'end_time': vparse_time,
            },
            'required_columns': (
                'trip_id',
                'start_time',
                'end_time',
                'headway_secs',
            ),
        }),
        ('routes.txt', {
            'converters': {
                'route_type': vparse_numeric,
            },
            'required_columns': (
                'route_id',
                'route_short_name',
                'route_long_name',
                'route_type',
            ),
        }),
        ('shapes.txt', {
            'converters': {
                'shape_pt_lat': vparse_numeric,
                'shape_pt_lon': vparse_numeric,
                'shape_pt_sequence': vparse_numeric,
                'shape_dist_traveled': vparse_numeric,
            },
            'required_columns': (
                'shape_id',
                'shape_pt_lat',
                'shape_pt_lon',
                'shape_pt_sequence',
            ),
        }),
        ('stops.txt', {
            'converters': {
                'stop_lat': vparse_numeric,
                'stop_lon': vparse_numeric,
                'location_type': vparse_numeric,
                'wheelchair_boarding': vparse_numeric,
                'pickup_type': vparse_numeric,
                'drop_off_type': vparse_numeric,
                'shape_dist_traveled': vparse_numeric,
                'timepoint': vparse_numeric,
            },
            'required_columns': (
                'stop_id',
                'stop_name',
                'stop_lat',
                'stop_lon',
            ),
        }),
        ('stop_times.txt', {
            'converters': {
                'arrival_time': vparse_time,
                'departure_time': vparse_time,
                'pickup_type': vparse_numeric,
                'shape_dist_traveled': vparse_numeric,
                'stop_sequence': vparse_numeric,
                'timepoint': vparse_numeric,
            },
            'required_columns': (
                'trip_id',
                'arrival_time',
                'departure_time',
                'stop_id',
                'stop_sequence',
            ),
        }),
        ('transfers.txt', {
            'converters': {
                'transfer_type': vparse_numeric,
                'min_transfer_time': vparse_numeric,
            },
            'required_columns': (
                'from_stop_id',
                'to_stop_id',
                'transfer_type',
            ),
        }),
        ('trips.txt', {
            'converters': {
                'direction_id': vparse_numeric,
                'wheelchair_accessible': vparse_numeric,
                'bikes_allowed': vparse_numeric,
            },
            'required_columns': (
                'route_id',
                'service_id',
                'trip_id',
            ),
        }),
    ])


'''
Writer configs
'''


def extract_agencies_config():
    G = empty_config()
    add_edge_config(G)

    G.remove_edges_from([
        ('routes.txt', 'trips.txt'),
        ('agency.txt', 'routes.txt'),
    ])

    G.add_edges_from([
        ('trips.txt', 'routes.txt', {
            'dependencies': {
                'route_id': 'route_id',
            },
        }),
        ('routes.txt', 'agency.txt', {
            'dependencies': {
                'agency_id': 'agency_id',
            },
        }),
    ])

    return G


def extract_routes_config():
    G = empty_config()
    add_edge_config(G)
    return G



import os
import shutil
import tempfile

from partridge.config import (
    default_config,
    extract_agencies_config,
    extract_routes_config,
)
from partridge.gtfs import feed as mkfeed


DEFAULT_NODES = frozenset(default_config().nodes())


def extract_agencies(inpath, outpath, agency_ids):
    config = extract_agencies_config()
    view = {'agency.txt': {'agency_id': agency_ids}}
    feed = mkfeed(inpath, config, view)
    return write_feed_dangerously(feed, outpath)


def extract_routes(inpath, outpath, route_ids):
    config = extract_routes_config()
    view = {'trips.txt': {'route_id': route_ids}}
    feed = mkfeed(inpath, config, view)
    return write_feed_dangerously(feed, outpath)


def write_feed_dangerously(feed, outpath, nodes=None):
    """
    Naively write a feed to a zipfile

    This function provides no sanity checks. Use it at
    your own risk.
    """
    nodes = DEFAULT_NODES if nodes is None else nodes
    try:
        tmpdir = tempfile.mkdtemp()

        for node in nodes:
            df = feed.get(node)
            if not df.empty:
                path = os.path.join(tmpdir, node)
                df.to_csv(path, index=False)

        if outpath.endswith('.zip'):
            outpath, _ = os.path.splitext(outpath)

        outpath = shutil.make_archive(outpath, 'zip', tmpdir)
    finally:
        shutil.rmtree(tmpdir)

    return outpath



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""The setup script."""
import io
import sys

from setuptools import setup, find_packages

with open('README.rst') as readme_file:
    readme = readme_file.read()

with open('HISTORY.rst') as history_file:
    history = history_file.read()

# About dict to store version and package info
about = dict()
with io.open('partridge/__version__.py', 'r', encoding='utf-8') as f:
    exec(f.read(), about)

requirements = [
    'networkx>=2.0',
    'pandas',
]

setup_requirements = [
    'pytest-runner',
]

test_requirements = [
    'pytest',
]

setup(
    name='partridge',
    version=about['__version__'],
    description='Partridge is python library for working with GTFS '
                'feeds using pandas DataFrames.',
    long_description=readme + '\n\n' + history,
    author='Danny Whalen',
    author_email='daniel.r.whalen@gmail.com',
    url='https://github.com/remix/partridge',
    packages=find_packages(include=['partridge']),
    include_package_data=True,
    install_requires=requirements,
    license='MIT license',
    zip_safe=False,
    keywords='partridge',
    classifiers=[
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Natural Language :: English',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
    ],
    python_requires='>=2.7, !=3.0.*, !=3.1.*, !=3.2.*, !=3.3.*, <4',
    test_suite='tests',
    tests_require=test_requirements,
    setup_requires=setup_requirements,
)



from datetime import datetime, date
from functools import partial
import numpy as np
import pandas as pd

DATE_FORMAT = '%Y%m%d'


def parse_time(val):
    if val is np.nan:
        return val

    val = val.strip()

    if val == '':
        return np.nan

    h, m, s = val.split(':')
    ssm = int(h) * 3600 + int(m) * 60 + int(s)

    # pandas doesn't have a NaN int, use floats
    return np.float64(ssm)


def parse_date(val):
    if isinstance(val, date):
        return val

    return datetime.strptime(val, DATE_FORMAT).date()


# Vectorized parse operations
vparse_date = np.vectorize(parse_date)
vparse_time = np.vectorize(parse_time)
vparse_numeric = partial(pd.to_numeric, errors='raise')
